@@ -2,14 +2,13 @@
 
 use std::cell::UnsafeCell;
 use std::io::{Error, Result, stderr};
-use std::os::unix::fs::{FileExt, FileTypeExt};
+use std::os::unix::fs::{FileExt, FileTypeExt, OpenOptionsExt};
 use std::os::unix::io::AsRawFd;
 use std::time::{Duration, Instant};
 use std::vec::Vec;
 
 use clap::Parser;
 use libc::{c_ushort, c_int, size_t};
-use nix::fcntl::{posix_fadvise, PosixFadviseAdvice};
 use nix::{ioctl_read, ioctl_read_bad, ioctl_write_ptr, request_code_none};
 use nix::unistd::isatty;
 use sensitive::alloc::Sensitive;
@@ -21,6 +20,10 @@ struct Opt {
 	#[clap(parse(from_os_str), value_hint = clap::ValueHint::FilePath)]
 	device: std::path::PathBuf,
 
+	/// Set idle I/O scheduling class
+	#[clap(short, long)]
+	idle: bool,
+
 	/// Do not overwrite corrupt sectors
 	#[clap(short('n'), long)]
 	dry_run: bool,
@@ -29,13 +32,14 @@ struct Opt {
 	#[clap(short, long)]
 	quiet: bool,
 
-	/// Write corrupted sector numbers to standard output
-	#[clap(long)]
-	sector_numbers: bool
+	/// Enumerate corrupt sectors to standard output
+	#[clap(short, long)]
+	enumerate: bool
 }
 
 struct Device {
-	file: std::fs::File,
+	direct: std::fs::File,
+	buffered: Option<std::fs::File>,
 	sectors: u64,
 	sector_size: usize,
 	maximum_io: u16,
@@ -76,30 +80,33 @@ struct Progress {
 
 ioctl_read_bad!(blksectget, request_code_none!(0x12, 103), c_ushort);
 ioctl_read_bad!(blksszget, request_code_none!(0x12, 104), c_int);
+ioctl_read!(blkbszget, 0x12, 112, size_t);
 ioctl_write_ptr!(blkbszset, 0x12, 113, size_t);
 ioctl_read!(blkgetsize64, 0x12, 114, u64);
 
 impl Device {
 	fn open<P: AsRef<std::path::Path>>(path: P, writable: bool) -> Result<Self> {
-		let file = std::fs::OpenOptions::new()
-			.read(true)
-			.write(writable)
-			.open(path)?;
+		let direct = std::fs::OpenOptions::new().read(true).custom_flags(libc::O_DIRECT).open(path)?;
+		let buffered = if writable {
+			Some(std::fs::OpenOptions::new().write(true).open(format!("/proc/self/fd/{}", direct.as_raw_fd()))?)
+		} else {
+			None
+		};
 
-		if !file.metadata()?.file_type().is_block_device() {
+		if !direct.metadata()?.file_type().is_block_device() {
 			use std::io::ErrorKind;
 			return Err(Error::new(ErrorKind::InvalidInput, "File is not a block device"));
 		}
 
 		let size = {
-			let mut size = u64::MAX;
-			unsafe { blkgetsize64(file.as_raw_fd(), &mut size) }?;
+			let mut size = 0;
+			unsafe { blkgetsize64(direct.as_raw_fd(), &mut size) }?;
 			size as u64
 		};
 
 		let sector_size = {
-			let mut ssz = -1;
-			unsafe { blksszget(file.as_raw_fd(), &mut ssz) }?;
+			let mut ssz = 0;
+			unsafe { blksszget(direct.as_raw_fd(), &mut ssz) }?;
 			assert!(ssz > 0);
 			ssz as usize
 		};
@@ -107,11 +114,24 @@ impl Device {
 		// Assert that device size is a multiple of the logical sector size
 		assert!(size % sector_size as u64 == 0);
 
-		unsafe { blkbszset(file.as_raw_fd(), &sector_size) }?;
+		let block_size = {
+			let mut bsz = 0;
+			unsafe { blkbszget(direct.as_raw_fd(), &mut bsz) }?;
+			assert!(bsz > 0);
+			bsz as usize
+		};
+
+		if block_size != sector_size {
+			unsafe { blkbszset(direct.as_raw_fd(), &sector_size) }?;
+
+			if let Some(ref file) = buffered {
+				unsafe { blkbszset(file.as_raw_fd(), &sector_size) }?;
+			}
+		}
 
 		let maximum_io = {
-			let mut sect = c_ushort::MAX;
-			unsafe { blksectget(file.as_raw_fd(), &mut sect) }?;
+			let mut sect = 0;
+			unsafe { blksectget(direct.as_raw_fd(), &mut sect) }?;
 			assert!(sect > 0);
 			sect as u16
 		};
@@ -121,14 +141,9 @@ impl Device {
 		// The allocator ensures that the memory is zero‐initialised
 		unsafe { buffer.set_len(maximum_io as usize * sector_size); }
 
-		for advice in [
-			PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
-			PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
-			posix_fadvise(file.as_raw_fd(), 0, 0, advice)?;
-		}
-
 		Ok(Self {
-			file,
+			direct,
+			buffered,
 			sectors: size / sector_size as u64,
 			sector_size,
 			maximum_io,
@@ -143,7 +158,7 @@ impl Device {
 		// The contents of this buffer are never examined
 		let buffer = unsafe { &mut *self.buffer.get() };
 
-		match self.file.read_at(&mut buffer[..count as usize * self.sector_size], offset * self.sector_size as u64) {
+		match self.direct.read_at(&mut buffer[..count as usize * self.sector_size], offset * self.sector_size as u64) {
 			Ok(len) => {
 				// Assert that we read a multiple of the sector size
 				assert!(len % self.sector_size == 0);
@@ -152,7 +167,7 @@ impl Device {
 			}
 
 			Err(err) => {
-				if let Some(libc::EIO) = err.raw_os_error() {
+				if let Some(libc::EILSEQ) = err.raw_os_error() {
 					Ok(None)
 				} else {
 					Err(err)
@@ -162,42 +177,30 @@ impl Device {
 	}
 
 	fn zero(&self, offset: u64) -> Result<()> {
-		let len = self.file.write_at(&self.null, offset * self.sector_size as u64)?;
-		assert!(len == self.sector_size);
-		Ok(())
+		self.buffered.as_ref().unwrap().write_all_at(&self.null, offset * self.sector_size as u64)
 	}
 
 	fn flush(&self, offset: u64, count: u16) -> Result<()> {
 		use libc::{sync_file_range, off64_t, SYNC_FILE_RANGE_WRITE};
 
-		if unsafe { sync_file_range(self.file.as_raw_fd(), (offset * self.sector_size as u64) as off64_t,
-			(count as usize * self.sector_size) as off64_t, SYNC_FILE_RANGE_WRITE) } != 0 {
-			Err(Error::last_os_error())
+		if let Some(ref file) = self.buffered {
+			if unsafe { sync_file_range(file.as_raw_fd(), (offset * self.sector_size as u64) as off64_t,
+				(count as usize * self.sector_size) as off64_t, SYNC_FILE_RANGE_WRITE) } != 0 {
+				Err(Error::last_os_error())
+			} else {
+				Ok(())
+			}
 		} else {
 			Ok(())
 		}
 	}
 
 	fn sync(&self) -> Result<()> {
-		self.file.sync_data()
-	}
-
-	fn acquire(&self, offset: u64, count: u16) -> Result<()> {
-		use libc::off_t;
-
-		posix_fadvise(self.file.as_raw_fd(), (offset * self.sector_size as u64) as off_t,
-		              (count as usize * self.sector_size) as off_t, PosixFadviseAdvice::POSIX_FADV_WILLNEED)?;
-
-		Ok(())
-	}
-
-	fn release(&self, offset: u64, count: u16) -> Result<()> {
-		use libc::off_t;
-
-		posix_fadvise(self.file.as_raw_fd(), (offset * self.sector_size as u64) as off_t,
-		              (count as usize * self.sector_size) as off_t, PosixFadviseAdvice::POSIX_FADV_DONTNEED)?;
-
-		Ok(())
+		if let Some(ref file) = self.buffered {
+			file.sync_data()
+		} else {
+			Ok(())
+		}
 	}
 
 	fn chunks(&self) -> u64 {
@@ -214,7 +217,6 @@ impl Device {
 
 impl Chunk<'_> {
 	fn iter(&self) -> SectorIterator {
-		self.device.acquire(self.index, self.count).unwrap();
 		SectorIterator {
 			chunk: self,
 			index: None
@@ -226,12 +228,6 @@ impl Chunk<'_> {
 	}
 }
 
-impl Drop for Chunk<'_> {
-	fn drop(&mut self) {
-		self.device.release(self.index, self.count).unwrap()
-	}
-}
-
 impl<'t> Iterator for ChunkIterator<'t> {
 	type Item = Result<Chunk<'t>>;
 
@@ -385,6 +381,12 @@ impl Progress {
 
 fn main() -> Result<()> {
 	let opt = Opt::parse();
+
+	if opt.idle {
+		use ioprio::{set_priority, Target, Pid, Priority, Class};
+		set_priority(Target::Process(Pid::this()), Priority::new(Class::Idle)).unwrap();
+	}
+
 	let dev = Device::open(&opt.device, !opt.dry_run)?;
 
 	let mut prog = Progress::new()?;
@@ -408,7 +410,7 @@ fn main() -> Result<()> {
 				if !sector.valid {
 					prog.error += 1;
 
-					if opt.sector_numbers {
+					if opt.enumerate {
 						println!("{}", sector.absolute());
 					}
 
